// SWPSetup01.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

namespace nSWPSQL
{
    static const CComBSTR bstrSetUpSQL[] =
    {
        L"CREATE TABLE SWPContractingJobs ("
            L"JobID INTEGER PRIMARY KEY IDENTITY(1, 1), "
            L"JobTitle VARCHAR(50) NOT NULL DEFAULT 'no title', "
            L"JobDescription VARCHAR(255) NOT NULL DEFAULT 'no description', "
            L"timestamp)",
        ////
        // Place Initial data before Trigger is created -- otherwise the
        // trigger fires on each insert
        L"INSERT INTO SWPContractingJobs "
            L"(JobTitle, JobDescription) VALUES "
            L"('Chief Technical Officer', 'SQL DB Engineer')",
        L"INSERT INTO SWPContractingJobs "
            L"(JobTitle, JobDescription) VALUES "
            L"('Architect', 'DNA and COM+ Czar')",
        L"INSERT INTO SWPContractingJobs "
            L"(JobTitle, JobDescription) VALUES "
            L"('Architect I', 'VB, C++ and Java')",
        L"INSERT INTO SWPContractingJobs "
            L"(JobTitle, JobDescription) VALUES "
            L"('Architect III', 'ASP.Net')",
        L"INSERT INTO SWPContractingJobs "
            L"(JobTitle, JobDescription) VALUES "
            L"('Team lead', 'ASP, MTS and MSMQ Wizard')",
        ////
        L"CREATE INDEX SWPWhoChanged "
        L"ON SWPContractingJobs (timestamp ASC)",
        ////
        L"CREATE INDEX SWPJobTitles "
        L"ON SWPContractingJobs (JobTitle ASC)",
        L"CREATE TRIGGER SWPTrigger01 ON SWPContractingJobs "
            L"AFTER INSERT, UPDATE AS "
            L"EXEC master..xp_cmdshell 'c:\\Bin\\SWPTrigger01.exe', NO_OUTPUT",
//        L"CREATE TRIGGER SWPTrigger01 ON SWPContractingJobs "
//            L"AFTER INSERT, UPDATE AS "
//            L"EXEC master..xp_cmdshell 'cscript c:\\Bin\\SWPTrigger01.wsf', NO_OUTPUT",

        // binary(8) is equivalent to timestamp
        L"CREATE TABLE SWPTriggerStamps ("
            L"TriggerName VARCHAR(128) PRIMARY KEY, "
            L"LastTimestamp binary(8) NOT NULL DEFAULT 0x00000000)",
        L"INSERT INTO SWPTriggerStamps "
            L"(TriggerName) VALUES ('SWPTrigger01')",
        // There is a chance that another trigger could update
        // before were are done.  For this reason the WHERE-clause
        // includes "@LastTimeStamp > LastTimeStamp"
        L"CREATE PROCEDURE SWPUpdateTriggerTimeStamp "
            L"@TriggerName VARCHAR(128), @LastTimeStamp VARBINARY(8) AS "
            L"UPDATE SWPTriggerStamps "
            L"SET LastTimeStamp = @LastTimeStamp "
            L"WHERE TriggerName = @TriggerName AND "
            L"@LastTimeStamp > LastTimeStamp",
        L""
    };

    static const CComBSTR bstrCleanUpSQL[] =
    {
        L"IF EXISTS (SELECT name FROM sysobjects "
                   L"WHERE name = 'SWPTrigger01' AND type = 'TR') "
        L"DROP Trigger SWPTrigger01",
        ////
        L"IF EXISTS (SELECT name FROM sysobjects "
                   L"WHERE name = 'SWPUpdateTriggerTimeStamp' AND type = 'P') "
        L"DROP PROCEDURE SWPUpdateTriggerTimeStamp",
        ////
        L"IF EXISTS (SELECT name FROM sysobjects "
        L"WHERE name = 'SWPTriggerStamps' AND type = 'U') "
        L"DROP TABLE SWPTriggerStamps",
        //// Drop table and associated indexes/triggers
        L"IF EXISTS (SELECT name FROM sysobjects "
        L"WHERE name = 'SWPContractingJobs' AND type = 'U') "
        L"DROP TABLE SWPContractingJobs",
        ////
        L""
    };

    static const CComBSTR bstrMasterSelect =
        L"SELECT JobTitle, JobDescription "
        L"FROM SWPContractingJobs "
        L"WHERE JobTitle LIKE '";
};



static HRESULT SWPExecute(CComPtr<ADODB::_Connection>& spConnection,
                          const CComBSTR bstrSQL[])

{
    HRESULT hr = S_OK;
    long lCount;
    long lLevel = 0;

    if (FAILED(hr = spConnection->BeginTrans(&lLevel)))
    {
        return hr;
    }

    for (lCount = 0; bstrSQL[lCount].Length(); lCount++)
    {
        if (FAILED(hr = spConnection->Execute(bstrSQL[lCount], 
                                              NULL,
                                              ADODB::adExecuteNoRecords | ADODB::adCmdText, 
                                              NULL)))
        {
            return hr;
        }
    }


    if (FAILED(hr = spConnection->CommitTrans()))
    {
        return hr;
    }

    return hr;
}


static HRESULT SWPGetNeverCachedData(CComPtr<ADODB::_Connection>& spConnection,
                                WCHAR whichLetter,
                                _bstr_t& bstrCachedData)

{
    HRESULT hr = S_OK;
    WCHAR letterArray[2] = {whichLetter, 0};
    CComBSTR bstrCommand = (WCHAR *)nSWPSQL::bstrMasterSelect;
    CComPtr<ADODB::_Recordset> spRecordset;
    CComPtr<ADODB::_Stream> spStream;
    long lSize;
    BSTR bstrTemp = NULL;

    if (FAILED(hr = spRecordset.CoCreateInstance(__uuidof(ADODB::Recordset))))
    {
        return hr;
    }

    bstrCommand += letterArray;
    bstrCommand += L"%'";
    if (FAILED(hr = spRecordset->Open(
                        CComVariant(bstrCommand),
                        CComVariant(spConnection),
                        ADODB::adOpenForwardOnly,
                        ADODB::adLockReadOnly,
                        ADODB::adCmdText)))
    {
        return hr;
    }

    if (FAILED(hr = spStream.CoCreateInstance(__uuidof(ADODB::Stream))))
    {
        return hr;
    }

    if (FAILED(hr = spRecordset->Save(CComVariant(spStream),
                                      ADODB::adPersistXML)))
    {
        return hr;
    }

    if (FAILED(hr = spStream->get_Size(&lSize)))
    {
        return hr;
    }

    if (FAILED(hr = spStream->ReadText(lSize, &bstrTemp)))
    {
        return hr;
    }

    bstrCachedData.Assign(bstrTemp);

    return hr;

} // SWPGetNeverCachedData()



static HRESULT SWPGetCachedData(CComPtr<ADODB::_Connection>& spConnection,
                                WCHAR whichLetter,
                                _bstr_t& bstrCachedData)

{
    HRESULT hr = S_OK;
    HANDLE hFile = NULL;
    DWORD dwSize = 0;
    DWORD dwNumRead = 0;
    string strCachedData; // data stored ini ANSI
    // Use _bstr_t because it converts WCHAR/CHAR automatically
    // The file name is the Table followed by the Letter queried
    _bstr_t bstrFileName = L"c:\\Bin\\SWPContractingJobs";
    WCHAR letterArray[2] = {whichLetter, 0};
    CComBSTR bstrCommand = (WCHAR *)nSWPSQL::bstrMasterSelect;

    bstrFileName += letterArray;
    bstrFileName += L".xml";
    bstrCommand += letterArray;
    bstrCommand += L"%'";

    // Loop here to avoid a race condition that takes place:
    // 1) if the table is updated
    // 2) the trigger files
    // 3) The trigger deletes the file just saved (spRecordset->Save)
    while (true)
    {
        if (!::PathFileExists((TCHAR *)bstrFileName))
        {
            CComPtr<ADODB::_Recordset> spRecordset;
            VARIANT vFileName;

            ::VariantInit(&vFileName);
            if (FAILED(hr = spRecordset.CoCreateInstance(__uuidof(ADODB::Recordset))))
            {
                return hr;
            }

            if (FAILED(hr = spRecordset->Open(
                                CComVariant(bstrCommand),
                                CComVariant(spConnection),
                                ADODB::adOpenForwardOnly,
                                ADODB::adLockReadOnly,
                                ADODB::adCmdText)))
            {
                return hr;
            }

            vFileName.vt = VT_BSTR;
            V_BSTR(&vFileName) = bstrFileName;
            if (FAILED(hr = spRecordset->Save(vFileName,
                                              ADODB::adPersistXML)))
            {
                return hr;
            }
        }

        hFile = ::CreateFile(bstrFileName,
                             GENERIC_READ,
                             FILE_SHARE_READ,
                             NULL,
                             OPEN_EXISTING,
                             0,
                             NULL);
        if (INVALID_HANDLE_VALUE == hFile)
        {
            DWORD dwError = ::GetLastError();

            if (ERROR_FILE_NOT_FOUND == dwError)
            {
                continue; // This handles the race condition
            }

            return HRESULT_FROM_WIN32(dwError);
        }

        if (0xFFFFFFFF == (dwSize = ::GetFileSize(hFile, NULL)))
        {
            hr = HRESULT_FROM_WIN32(::GetLastError());
            break;
        }

        strCachedData.resize(dwSize);
        if (0 == strCachedData.length())
        {
            hr = E_OUTOFMEMORY;
            break;
        }

        if (!::ReadFile(hFile, 
                        const_cast<CHAR *>(strCachedData.c_str()), 
                        dwSize, 
                        &dwNumRead, 
                        NULL))
        {
            hr = HRESULT_FROM_WIN32(::GetLastError());
            break;
        }

        break;
    }

    if (!::CloseHandle(hFile))
    {
        // hr may have been set to error in the while loop
        if (SUCCEEDED(hr))
        {
            hr = HRESULT_FROM_WIN32(::GetLastError());
        }
    }

    bstrCachedData = strCachedData.c_str();

    return hr;

} // SWPGetCachedData()



static HRESULT SWPCacheDemo(CComPtr<ADODB::_Connection>& spConnection)

{
    HRESULT hr = S_OK;
    string strData;
    CComBSTR bstrCommand;
    _bstr_t bstrCachedData;

    for (long lCount = 0; lCount < 5; lCount++)
    {
        if (!((lCount+1) % 3)) // 0th, 1st then 2nd time through edit data 
        {                      // trigger will force cache update
            bstrCommand = L"INSERT INTO SWPContractingJobs "
                L"(JobTitle, JobDescription) VALUES "
                L"('Architect XV', 'Windows Developer, 15th level black belt')";
            if (FAILED(hr = spConnection->Execute(bstrCommand, 
                                                  NULL,
                                                  ADODB::adExecuteNoRecords | ADODB::adCmdText, 
                                                  NULL)))
            {
                return hr;
            }
        }

        if (FAILED(SWPGetCachedData(spConnection, L'A', bstrCachedData)))
        {
            return hr;
        }

        ::MessageBox(NULL, 
                     bstrCachedData, 
                     _T("Cache file: c:\\Bin\\SWPContractingJobs<letter>.xml"), 
                     MB_OK);
    }

    return hr;
}



static HRESULT SWPShowTiming(CComPtr<ADODB::_Connection>& spConnection)
{
    _bstr_t bstrCachedData;
    long lTimingCount = 0;
    LARGE_INTEGER liStart; 
    LARGE_INTEGER liEnd; 
    LARGE_INTEGER liFrequency;
    static const long lMaxCount = 1000;
    FLOAT fTime;
    HRESULT hr = S_OK;
    DWORD dwOrgAffinityMask = 0;

    if (!(dwOrgAffinityMask = 
          ::SetThreadAffinityMask(::GetCurrentThread(), 1)))
    {
        return HRESULT_FROM_WIN32(::GetLastError());
    }

    ::QueryPerformanceFrequency(&liFrequency);
    ::Sleep(0);
    ::QueryPerformanceCounter(&liStart);
    for (lTimingCount = 0; lTimingCount < lMaxCount; lTimingCount++)
    {
        if (FAILED(SWPGetCachedData(spConnection, L'A', bstrCachedData)))
        {
            return hr;
        }
    }

    ::QueryPerformanceCounter(&liEnd);
    fTime = (FLOAT)(liEnd.QuadPart - liStart.QuadPart)/
                   (FLOAT)(liFrequency.QuadPart);
    cout << "Time for (" << lMaxCount << ") cached iterations:  " << fTime << endl;

    ::Sleep(0);
    ::QueryPerformanceCounter(&liStart);
    for (lTimingCount = 0; lTimingCount < lMaxCount; lTimingCount++)
    {
        if (FAILED(SWPGetNeverCachedData(spConnection, L'A', bstrCachedData)))
        {
            return hr;
        }
    }

    ::QueryPerformanceCounter(&liEnd);
    fTime = (FLOAT)(liEnd.QuadPart - liStart.QuadPart)/
                   (FLOAT)(liFrequency.QuadPart);
    cout << "Time for (" << lMaxCount << ") non-cached iterations: " << fTime << endl;

    if (!::SetThreadAffinityMask(::GetCurrentThread(), dwOrgAffinityMask))
    {
        return HRESULT_FROM_WIN32(::GetLastError());
    }

    return hr;
}



static HRESULT SWPRunIt(bool bOnlyCreateData)

{
    HRESULT hr = S_OK;
    CComPtr<ADODB::_Connection> spConnection;
    CComBSTR bstrConnStr = 
        L"PROVIDER=SQLOLEDB;DATA SOURCE=localhost;"
        L"DATABASE=SWPASPToday;USER ID=sa;PASSWORD=sa;";

    if (FAILED(hr = 
        spConnection.CoCreateInstance(__uuidof(ADODB::Connection))))
    {
        return hr;
    }

    if (FAILED(hr = spConnection->Open(bstrConnStr,
                                       NULL,
                                       NULL,
                                       ADODB::adConnectUnspecified)))
    {
        return hr;
    }

    if (FAILED(hr = SWPExecute(spConnection,
                               nSWPSQL::bstrCleanUpSQL)))
    {
        return hr;
    }

    if (FAILED(hr = SWPExecute(spConnection,
                               nSWPSQL::bstrSetUpSQL)))
    {
        return hr;
    }

// This code demonstrates that retrieving from a file is faster than
// an ADO/SQL query against SQL Server
//    if (FAILED(hr = SWPShowTiming(spConnection)))
//    {
//        return hr;
//    }

    if (!bOnlyCreateData)
    {
        if (FAILED(hr = SWPCacheDemo(spConnection)))
        {
            return hr;
        }
    }

    if (FAILED(hr = spConnection->Close()))
    {
        return hr;
    }

    spConnection = NULL;

    return hr;
}



int main(int argc, char* argv[])
{
    HRESULT hr = S_OK;

    if (FAILED(hr = ::CoInitialize(NULL)))
    {
        return hr;
    }

    // If there are no command-line args (argc==1) then we
    // only create the data.  If there is at least one 
    // command-line arg then we create the data and demonstrate
    // cache.
    if (FAILED(hr = SWPRunIt(1 == argc)))
    {
        return hr;
    }

    ::CoUninitialize();

    return 0;
}

