// This is needed for MessageBox's MB_SERVER_NOTIFICATION flag
#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0500
#endif

#include <windows.h>
#include <tchar.h>
#pragma warning (disable: 4146) // ADO generated signed/unsigned warning
#import <msado15.dll> raw_interfaces_only
#include <atlbase.h>


 
static HRESULT SWPUpdateTriggerCounter(CComPtr<ADODB::_Connection>& spConnection,
                                       const CComVariant& vTriggername,
                                       const CComVariant& vMaxTimeStamp)

{
    HRESULT hr = S_OK;
    CComPtr<ADODB::_Command> spCommand;
    CComPtr<ADODB::Parameters> spParameters;
    CComPtr<ADODB::_Parameter> spParmTrigName, spParmLastTimeStamp;
    CComBSTR bstrSelectCommand = L"SWPUpdateTriggerTimeStamp";
    VARIANT vConnection; // do not clear this variant since 
                         // connection is a CComPtr

    ::VariantInit(&vConnection);
    V_DISPATCH(&vConnection) = spConnection;
    vConnection.vt = VT_DISPATCH;
    if (FAILED(hr = spCommand.CoCreateInstance(__uuidof(ADODB::Command))))
    {
        return hr;
    }

    if (FAILED(hr = spCommand->put_ActiveConnection(vConnection)))
    {
        return hr;
    }

    if (FAILED(hr = spCommand->put_CommandText(bstrSelectCommand)))
    {
        return hr;
    }

    if (FAILED(hr = spCommand->put_CommandType(ADODB::adCmdStoredProc)))
    {
        return hr;
    }

    if (FAILED(hr = spCommand->get_Parameters(&spParameters)))
    {
        return hr;
    }

    if (FAILED(hr = spParameters->get_Item(CComVariant(CComBSTR(L"@TriggerName")), 
                                           &spParmTrigName)))
    {
        return hr;
    }

    if (FAILED(hr = spParameters->get_Item(CComVariant(CComBSTR(L"@LastTimeStamp")), 
                                           &spParmLastTimeStamp)))
    {
        return hr;
    }

    if (FAILED(hr = spParmTrigName->put_Value(vTriggername)))
    {
        return hr;
    }

    if (FAILED(hr = spParmLastTimeStamp->put_Value(vMaxTimeStamp)))
    {
        return hr;
    }

    if (FAILED(hr = spCommand->Execute(NULL, NULL, ADODB::adExecuteNoRecords, NULL)))
    {
        return hr;
    }

    return hr;

} // SWPUpdateTriggerCounter()



static HRESULT SWPProcessTrigger()

{
    HRESULT hr = S_OK;
    CComPtr<ADODB::_Connection> spConnection;
    CComBSTR bstrConnStr = 
        L"PROVIDER=SQLOLEDB;DATA SOURCE=localhost;"
        L"DATABASE=SWPASPToday;USER ID=sa;PASSWORD=sa;";

    if (FAILED(hr = 
        spConnection.CoCreateInstance(__uuidof(ADODB::Connection))))
    {
        return hr;
    }

    if (FAILED(hr = spConnection->Open(bstrConnStr,
                                       NULL,
                                       NULL,
                                       ADODB::adConnectUnspecified)))
    {
        return hr;
    }

    CComVariant vLetterValue, vTimeStampValue, vMaxTimeStamp;
    CComPtr<ADODB::_Recordset> spRecordset;
    CComPtr<ADODB::Fields> spFields;
    CComPtr<ADODB::Field> spFieldJobTitleLetter, spFieldTimeStamp;
    CComBSTR bstrSelectCommand =
        L"SELECT LEFT(A.JobTitle, 1) JobTitleLetter, "
               L"MAX(A.timestamp) LetterTimeStamp "
        L"FROM SWPContractingJobs A, SWPTriggerStamps B "
        L"WHERE (A.timestamp>B.LastTimeStamp) AND "
              L"(B.TriggerName='SWPTrigger01') "
        L"GROUP BY LEFT(A.JobTitle, 1)";
    bool bValueGTMax;

    if (FAILED(hr = spConnection->Execute(bstrSelectCommand, 
                                          NULL, 
                                          ADODB::adCmdText, 
                                          &spRecordset)))
    {
        return hr;
    }

    if (FAILED(hr = spRecordset->get_Fields(&spFields)))
    {
        return hr;
    }

    if (FAILED(hr = spFields->get_Item(CComVariant(CComBSTR(L"JobTitleLetter")),
                                       &spFieldJobTitleLetter)))
    {
        return hr;
    }

    if (FAILED(hr = spFields->get_Item(CComVariant(CComBSTR(L"LetterTimeStamp")),
                                       &spFieldTimeStamp)))
    {
        return hr;
    }

    VARIANT_BOOL vbEOF = VARIANT_FALSE;
    // _bstr_t converts from ANSI to/from Unicode so it
    // works with ANSI/Unicode Win32 API's such as DeleteFile
    _bstr_t bstrFileName;
    DWORD dwError = ERROR_SUCCESS;

    while (true)
    {
        if (FAILED(hr = spRecordset->get_EOF(&vbEOF)))
        {
            return hr;
        }

        if (VARIANT_TRUE == vbEOF)
        {
            break;
        }

        vLetterValue.Clear();
        if (FAILED(hr = spFieldJobTitleLetter->get_Value(&vLetterValue)))
        {
            return hr;
        }


        // create new XML cache
        bstrFileName = L"c:\\Bin\\SWPContractingJobs";
        bstrFileName += reinterpret_cast<WCHAR *>V_BSTR(&vLetterValue);
        bstrFileName += L".xml";
        while (!::DeleteFile(bstrFileName))
        {
            dwError = ::GetLastError();
            if (ERROR_FILE_NOT_FOUND == dwError)
            {
                break;
            }

            else if (ERROR_SHARING_VIOLATION != dwError)
            {
                return HRESULT_FROM_WIN32(dwError);
            }

            ::Sleep(0); // Surrender time quantum
        }

        vTimeStampValue.Clear();
        if (FAILED(hr = spFieldTimeStamp->get_Value(&vTimeStampValue)))
        {
            return hr;
        }

        if (VT_EMPTY == vMaxTimeStamp.vt)
        {
            vMaxTimeStamp = vTimeStampValue;
        }

        else
        {
            VARIANT_BOOL *pbMax, *pbValue;

            if (FAILED(hr = ::SafeArrayAccessData(V_ARRAY(&vMaxTimeStamp), 
                                  reinterpret_cast<VOID **>(&pbMax))))
            {
                return hr;
            }

            if (FAILED(hr = ::SafeArrayAccessData(V_ARRAY(&vTimeStampValue), 
                                  reinterpret_cast<VOID **>(&pbValue))))
            {
                return hr;
            }

            bValueGTMax = (0 < memcmp(pbValue, pbMax, 8));
            if (FAILED(hr = ::SafeArrayUnaccessData(V_ARRAY(&vMaxTimeStamp))))
            {
                return hr;
            }

            if (FAILED(hr = ::SafeArrayUnaccessData(V_ARRAY(&vTimeStampValue))))
            {
                return hr;
            }

            if (bValueGTMax)
            {
                vMaxTimeStamp = vTimeStampValue;
            }

        }

        if (FAILED(hr = spRecordset->MoveNext()))
        {
            return hr;
        }
    }

    // If empty then there was nothing to update.  This is possible because
    // a trigger fired at the same time could have already done our work for us
    if (VT_EMPTY != vMaxTimeStamp.vt)
    {
        if (FAILED(hr = SWPUpdateTriggerCounter(spConnection,
                                                CComVariant(CComBSTR("SWPTrigger01")),
                                                vMaxTimeStamp)))
        {
            return hr;
        }
    }

    if (FAILED(hr = spConnection->Close()))
    {
        return hr;
    }

    spConnection = NULL;

    return hr;

} // SWPProcessTrigger()



int main(int argc, char* argv[])
{
    HRESULT hr = S_OK;

// Uncomment this to run trigger in the debugger
//    ::MessageBox(NULL, _T("Text"), _T("Caption"), MB_SERVICE_NOTIFICATION);

    if (1 == argc)
    {
        // Use this because it support Unicode/ANSI
        _bstr_t bstrProcess = argv[0];
        STARTUPINFO si = {{0}};
        PROCESS_INFORMATION pi = {{0}};

        bstrProcess += L" AnyParm";
        si.cb = sizeof si;
        if (!::CreateProcess(NULL,
                             bstrProcess,
                             NULL,
                             NULL,
                             FALSE,
                             CREATE_NO_WINDOW,
                             NULL,
                             NULL,
                             &si,
                             &pi))
        {
            return HRESULT_FROM_WIN32(::GetLastError());
        }

        else
        {
            return 0;
        }
    }

    if (FAILED(hr = ::CoInitialize(NULL)))
    {
        return hr;
    }

    if (FAILED(hr = SWPProcessTrigger()))
    {
        return hr;
    }

    ::CoUninitialize();

    return 0;

} // main()

